//-----------------------------------------------------------------------------
#include <stdio.h>
#include <iostream>
#include <mvIMPACT_CPP/mvIMPACT_acquire.h>
#include "saveimg.h"
#ifdef MALLOC_TRACE
	#include <mcheck.h>
#endif	// MALLOC_TRACE

using namespace mvIMPACT::acquire;
using namespace std;

//-----------------------------------------------------------------------------
Device* getDeviceFromUserInput( const DeviceManager& devMgr )
//-----------------------------------------------------------------------------
{
	unsigned int devCnt = devMgr.deviceCount();
	if( devCnt == 0 )
	{
		cout << "*** No MATRIX VISION device found!" << endl;
		return 0;
	}

	// display every device detected
	for( unsigned int i=0; i<devCnt; i++ )
	{
		Device* pDev = devMgr[i];
		if( pDev )
		{
			cout << "[" << i << "]: " << pDev->serial.read() << " (" << pDev->product.read() << ", family: " << pDev->family.read() << ")" << endl;
		}
	}

	// get user input
	cout << endl << "Please enter the number in front of the listed device followed by [ENTER] to open it: ";
	unsigned int devNr = 0;
	cin >> devNr;
	// remove the '\n' from the stream
	cin.get();

	if( devNr >= devCnt )
	{
		cout << "*** Invalid selection!" << endl;
		return 0;
	}

	cout << "Using device number " << devNr << ".(" << devMgr[devNr]->serial.read() << ")" << endl;
	return devMgr[devNr];
}

//-----------------------------------------------------------------------------
int main( int /*argc*/, char* /*argv*/[])
//-----------------------------------------------------------------------------
{
#ifdef MALLOC_TRACE
	mtrace();
#endif	// MALLOC_TRACE

	cout << endl << " ++ Start sample: " << __DATE__ << "/" << __TIME__ << endl;

	DeviceManager devMgr;
	Device* pDev = getDeviceFromUserInput( devMgr );

	if( pDev )
	{
		try
		{
			pDev->open();
		}
		catch( ImpactAcquireException& e )
		{
			// this e.g. might happen if the same device is already opened in another process...
			cout << "*** Error: An error occurred while opening the device (error: " << e.getErrorString() << "! code: " << e.getErrorCode() << ")." << endl;
			cout << "Press RETURN to end the application..." << endl;
			cin.get();
			return 0;
		}
	}
	else
	{
		cout << "Press RETURN to end the application..." << endl;
		cin.get();
		return 0;
	}

	// only 8Bit/pixel destination image are supported by the writeFile() function
	ImageDestination imgDst(pDev);
	imgDst.pixelFormat.write( idpfMono8 );

	CameraSettingsBase cs(pDev);
	cs.imageRequestTimeout_ms.write( 10000 ); // e.g. USB 1.1 on an embedded system needs a large timeout for the first image

	FunctionInterface fi(pDev);
	// make sure there is at least 1 request to capture an image
	// requests are structures containing information for the driver how to capture an image
	// and return information about the image back to user when the image has been captured
	fi.ensureRequests( 1 );
	// send a request to the default request queue of the device and wait for the result.
	fi.imageRequestSingle();

	// wait for results from the default capture queue
	int iRequestNr = fi.imageRequestWaitFor( -1 );

	cout << "Request Nr.: " << iRequestNr << endl;
	// check if the image has been captured without any problems
	if( !fi.isRequestNrValid( iRequestNr ) )
	{
		// this can not happen in this sample, but may happen if you wait for a request without
		// sending one to the driver before
		cout << "*** Error: " << ImpactAcquireException::getErrorCodeAsString( iRequestNr ) << "(" << iRequestNr << ")" << endl;
		return 0;
	}

	const Request* pRequest = fi.getRequest(iRequestNr);
	if( !fi.isRequestOK( pRequest ) )
	{
		cout << "*** Error: " << pRequest->requestResult.readS() << endl;
		return 0;
	}

	// everything went well. Store the result into a file
	const string sImageName = string("SingCapt.pgm");
	cout << "Will save the file as:  " << sImageName << endl;
	if( writeFile( (const unsigned char*)pRequest->imageData.read(), pRequest->imageWidth.read(), pRequest->imageHeight.read(), pRequest->imageLinePitch.read(), pRequest->imagePixelPitch.read(), sImageName.c_str() ) < 0 )
	{
		cout << "*** Error: File: "<< sImageName << " could not be saved!!" << endl;
	}

	// unlock the buffer to let the driver know that you no longer need this buffer
	fi.imageRequestUnlock( iRequestNr );

	cout << "Press RETURN to end the application" << endl;
	cin.get();

	// free resources
	fi.imageRequestReset( 0, 0 );

	pDev->close();

	return 0;
}
