#include "ImageCanvas.h"
#include "PropViewFrame.h"
#include <vector>

IMPLEMENT_DYNAMIC_CLASS(ImageCanvas, DrawingCanvas)

BEGIN_EVENT_TABLE(ImageCanvas, DrawingCanvas)
  EVT_PAINT(ImageCanvas::OnPaint)
  EVT_KEY_DOWN(ImageCanvas::OnKeyDown)
  EVT_LEFT_DCLICK(ImageCanvas::OnDblClick)
END_EVENT_TABLE()

using namespace std;

//-----------------------------------------------------------------------------
std::vector<wxString>::size_type Split( const wxString& str, const wxString& separator, std::vector<wxString>& v )
//-----------------------------------------------------------------------------
{
	v.clear();
	if( str.IsEmpty() )
	{
		return 0;
	}

	wxString tokens(str);
	int posSep = 0;
	while( ( posSep = tokens.Find( separator.c_str() ) ) != -1 )
	{
		if( posSep == 0 )
		{
			// starts with a separator
			tokens = tokens.Mid( posSep + separator.Len() );
			continue;
		}
		v.push_back( tokens.Mid( 0, posSep ) );
		tokens = tokens.Mid( posSep + separator.Len() );
	}
	if( !tokens.IsEmpty() )
	{
		v.push_back( tokens );
	}
	return v.size();
}

//-----------------------------------------------------------------------------
inline long GetNextLongParam( const wxString& token )
//-----------------------------------------------------------------------------
{
	long val = 0;
	token.ToLong( &val );
	return val;
}

//-----------------------------------------------------------------------------
inline double GetNextDoubleParam( const wxString& token )
//-----------------------------------------------------------------------------
{
	double val = 0;
	token.ToDouble( &val );
	return val;
}

//-----------------------------------------------------------------------------
void ImageCanvas::BlitAOIs( wxPaintDC& dc, double scaleFactor, int bmpXOff, int bmpYOff, int bmpW, int bmpH )
//-----------------------------------------------------------------------------
{
	dc.SetBrush( *wxTRANSPARENT_BRUSH );
	AOISet::iterator it = m_aoiSet.begin();
	AOISet::const_iterator itEnd = m_aoiSet.end();
	while( it != itEnd )
	{
		dc.SetPen( (*it)->m_colour );
		int aoix = (*it)->m_rect.GetX();
		int aoiy = (*it)->m_rect.GetY();
		int aoiw = (*it)->m_rect.GetWidth();
		int aoih = (*it)->m_rect.GetHeight();
		dc.DrawRectangle( static_cast<int>(( aoix * scaleFactor ) + bmpXOff),
			static_cast<int>(( aoiy * scaleFactor ) + bmpYOff),
			( aoiw + aoix > bmpW ) ? static_cast<int>(( bmpW - aoix ) * scaleFactor) : static_cast<int>(aoiw * scaleFactor),
			( aoih + aoiy > bmpH ) ? static_cast<int>(( bmpH - aoiy ) * scaleFactor) : static_cast<int>(aoih * scaleFactor) );
		dc.SetTextForeground( (*it)->m_colour );
		wxCoord textWidth, textHeight;
		dc.GetTextExtent( (*it)->m_description, &textWidth, &textHeight );
		dc.DrawText( (*it)->m_description,
			static_cast<wxCoord>(( aoix * scaleFactor ) + bmpXOff + ( ( static_cast<int>(aoiw * scaleFactor) - textWidth ) / 2 )),
			static_cast<wxCoord>(( aoiy * scaleFactor ) + bmpYOff + ( ( static_cast<int>(aoih * scaleFactor) - textHeight ) / 2 )) );
		dc.SetPen( wxNullPen );
		++it;
	}
}

//-----------------------------------------------------------------------------
/// Strings containing \a _PVDI_ (PropView Display Info) are treated differently.
///
/// These strings can be used to blit special information on top of the image.
/// Supported things are strings rectangles and ellipse overlays.
///
/// The string must have the following format:
///
/// token;token;token;token
///
/// one of these tokens must be \a _PVDI_ to inform the display that this string has to be treated
/// in a special way. Other valid tokens are:
///
/// - _RECT_=top,left,bottom,right,r,g,b
/// - _ELLI_=x,y,w,h,r,g,b
/// - _STR_=text,x,y,r,g,b
/// - _LN_=x1,y1,x2,y2,r,g,b (line)
///
/// _PVDI_;_STR_=text,x,y,r,g,b;_RECT_=top,left,bottom,right,r,g,b;_ELLI_=x,y,w,h,r,g,b
///
/// EXAMPLES:
///
/// _PVDI_;_STR_=dummy,100,100,0,255,0;_RECT_=50,50,200,200,255,0,0
///
/// This will draw 'dummy' at 100/100 in green on top of the image and a red rectangle from 50/50 to 200/200
///
/// _PVDI_;_STR_=dummy,100,100;_ELLI_=50,50,200,200,255,0,0
///
/// This will draw a green 'dummy' at 100/100(no color info -> default color) and a red circle (as w and h are equal)
/// at 50/50 with a radius of 200. Color information can be omitted. Then 0/255/0 (green) will be used.
void ImageCanvas::BlitInfoStrings( wxPaintDC& dc, double scaleFactor, int bmpXOff, int bmpYOff, int bmpW, int bmpH )
//-----------------------------------------------------------------------------
{
	if( m_boShowInfoOverlay )
	{
		wxColour oldTextColor(dc.GetTextForeground());
		std::vector<wxString>::size_type vSize = m_infoStringsOverlay.size();
		for( std::vector<wxString>::size_type i=0; i<vSize; i++ )
		{
			dc.SetTextForeground( wxColour(0, 255, 0) );
			dc.SetPen( wxColour(0, 255, 0) );
			static const wxString SPECIAL_DISPLAY_INFO_TOKEN(wxT("_PVDI_"));
			int displayInfoPos = m_infoStringsOverlay[i].Find( SPECIAL_DISPLAY_INFO_TOKEN.c_str() );
			if( displayInfoPos >= 0 )
			{
				int pos = m_infoStringsOverlay[i].Find( wxT(":") );
				if( pos >= 0 )
				{
					vector<wxString> tokens;
					wxString tokenList(m_infoStringsOverlay[i].Mid( pos+1 ));
					tokenList.Trim().Trim( false );
					vector<wxString>::size_type tokenCnt = Split( tokenList, wxString(wxT(";")), tokens );
					for( vector<wxString>::size_type j=0; j<tokenCnt; j++ )
					{
						vector<wxString> params;
						vector<wxString>::size_type paramsCnt = Split( tokens[j], wxString(wxT("=")), params );
						if( ( ( paramsCnt == 1 ) && ( tokens[j] != SPECIAL_DISPLAY_INFO_TOKEN ) ) /*|| ( ( paramsCnt < 1 ) || ( paramsCnt > 2 ) )*/ )
						{
							dc.DrawText( wxString::Format( wxT("Invalid parameter string: %s"), m_infoStringsOverlay[i].c_str() ), bmpXOff, bmpYOff + 300 + (i*20) );
							break;
						}

						if( params[0] == SPECIAL_DISPLAY_INFO_TOKEN )
						{
							continue;
						}

						vector<wxString> data;
						vector<wxString>::size_type dataCnt = Split( params[1], wxString(wxT(",")), data );
						dc.SetPen( wxColour(0, 255, 0) ); // the default pen
						dc.SetTextForeground( wxColour(0, 255, 0) ); // the default text colour
						if( params[0] == wxT("_STR_") )
						{
							if( dataCnt < 3 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 6 )
							{
								dc.SetTextForeground( wxColour(GetNextLongParam( data[3] ), GetNextLongParam( data[4] ), GetNextLongParam( data[5] )) );
							}
							long x = GetNextLongParam( data[1] );
							long y = GetNextLongParam( data[2] );
							dc.DrawText( data[0], static_cast<wxCoord>(( x * scaleFactor ) + bmpXOff ), static_cast<wxCoord>(( y * scaleFactor ) + bmpYOff ));
						}
						else if( params[0] == wxT("_RECT_") )
						{
							if( dataCnt < 4 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 7 )
							{
								dc.SetPen( wxColour(GetNextLongParam( data[4] ), GetNextLongParam( data[5] ), GetNextLongParam( data[6] )) );
							}
							long y = GetNextLongParam( data[0] );
							long x = GetNextLongParam( data[1] );
							long bottom = GetNextLongParam( data[2] );
							long right = GetNextLongParam( data[3] );
							dc.DrawRectangle( static_cast<int>(( x * scaleFactor ) + bmpXOff),
								static_cast<int>(( y * scaleFactor ) + bmpYOff),
								( right > bmpW ) ? static_cast<int>(( bmpW - x ) * scaleFactor) : static_cast<int>(( right - x ) * scaleFactor),
								( bottom > bmpH ) ? static_cast<int>(( bmpH - y ) * scaleFactor) : static_cast<int>(( bottom - y ) * scaleFactor) );
						}
						else if( params[0] == wxT("_ELLI_") )
						{
							if( dataCnt < 4 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 7 )
							{
								dc.SetPen( wxColour(GetNextLongParam( data[4] ), GetNextLongParam( data[5] ), GetNextLongParam( data[6] )) );
							}
							long x = GetNextLongParam( data[0] );
							long y = GetNextLongParam( data[1] );
							long w = GetNextLongParam( data[2] );
							long h = GetNextLongParam( data[3] );
							dc.DrawEllipse( static_cast<int>(( x * scaleFactor ) + bmpXOff),
								static_cast<int>(( y * scaleFactor ) + bmpYOff),
								( w + x > bmpW ) ? static_cast<int>(( bmpW - x ) * scaleFactor) : static_cast<int>(w * scaleFactor),
								( h + y > bmpH ) ? static_cast<int>(( bmpH - y ) * scaleFactor) : static_cast<int>(h * scaleFactor) );
						}
						else if( params[0] == wxT("_LN_") )
						{
							if( dataCnt < 4 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 7 )
							{
								dc.SetPen( wxColour(GetNextLongParam( data[4] ), GetNextLongParam( data[5] ), GetNextLongParam( data[6] )) );
							}
							long x1 = GetNextLongParam( data[0] );
							long y1 = GetNextLongParam( data[1] );
							long x2 = GetNextLongParam( data[2] );
							long y2 = GetNextLongParam( data[3] );
							dc.DrawLine( static_cast<int>(( x1 * scaleFactor ) + bmpXOff),
								static_cast<int>(( y1 * scaleFactor ) + bmpYOff),
								static_cast<int>(( x2 * scaleFactor ) + bmpXOff),
								static_cast<int>(( y2 * scaleFactor ) + bmpYOff) );
						}
						else if( params[0] == wxT("_PT_") )
						{
							if( dataCnt < 2 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 5 )
							{
								dc.SetPen( wxColour(GetNextLongParam( data[2] ), GetNextLongParam( data[3] ), GetNextLongParam( data[4] )) );
							}
							long x = GetNextLongParam( data[0] );
							long y = GetNextLongParam( data[1] );
							dc.DrawPoint( static_cast<int>(( x * scaleFactor ) + bmpXOff),
								static_cast<int>(( y * scaleFactor ) + bmpYOff) );
						}
						else if( params[0] == wxT("_DPT_") )
						{
							if( dataCnt < 2 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 5 )
							{
								dc.SetPen( wxColour(GetNextLongParam( data[2] ), GetNextLongParam( data[3] ), GetNextLongParam( data[4] )) );
							}
							double x = GetNextDoubleParam( data[0] );
							double y = GetNextDoubleParam( data[1] );
							dc.DrawPoint( static_cast<int>(( x * scaleFactor ) + bmpXOff),
								static_cast<int>(( y * scaleFactor ) + bmpYOff) );
						}
						else if( params[0] == wxT("_3DPT_") )
						{
							if( dataCnt < 2 )
							{
								dc.SetTextForeground( wxColour(255, 0, 0) );
								dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
								continue;
							}
							if( dataCnt >= 6 )
							{
								dc.SetPen( wxColour(GetNextLongParam( data[3] ), GetNextLongParam( data[4] ), GetNextLongParam( data[5] )) );
							}
							double x = GetNextDoubleParam( data[0] );
							double y = GetNextDoubleParam( data[1] );
							dc.DrawPoint( static_cast<int>(( x * scaleFactor ) + bmpXOff),
								static_cast<int>(( y * scaleFactor ) + bmpYOff) );
						}
						else
						{
							dc.SetTextForeground( wxColour(255, 0, 0) );
							dc.DrawText( wxString::Format( wxT("Invalid token: %s"), tokens[j].c_str() ), bmpXOff, bmpYOff + 320 + (i*20) );
						}
					}
				}
				else
				{
					dc.DrawText( m_infoStringsOverlay[i], bmpXOff, bmpYOff + INFO_Y_OFFSET + (i*20) );
				}
			}
			else
			{
				dc.DrawText( m_infoStringsOverlay[i], bmpXOff, bmpYOff + INFO_Y_OFFSET + (i*20) );
			}
		}
		dc.SetPen( wxNullPen );
		dc.SetTextForeground( oldTextColor );
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::BlitPerformanceMessages( wxPaintDC& dc, int bmpXOff, int bmpYOff, TImageBufferPixelFormat pixelFormat )
//-----------------------------------------------------------------------------
{
	if( m_boShowOverlayMessages )
	{
		bool boFormatWarning = ( pixelFormat != ibpfMono8 ) && ( pixelFormat != ibpfRGBx888Packed );
		dc.SetTextForeground( wxColour(255, 0, 0) );
		wxString msg(wxT("Performance loss because of"));
		bool boMustShow = false;
		if( boFormatWarning )
		{
			msg.Append( wxT(" pixel format conversion for display") );
			boMustShow = true;
		}

		if( m_boScaleToClientSize )
		{
			if( boMustShow )
			{
				msg.Append( wxT(" and") );
			}
			msg.Append( wxT(" scaling to fit to client size") );
			boMustShow = true;
		}

		if( !m_boScaleToClientSize && ( m_currentZoomFactor != 1.0 ) )
		{
			if( boMustShow )
			{
				msg.Append( wxT(" and") );
			}
			msg.Append( wxString::Format( wxT(" scaling(approx. zoom factor: %.1f)"), m_currentZoomFactor ) );
			boMustShow = true;
		}

		if( boMustShow )
		{
			dc.DrawText( msg, bmpXOff, bmpYOff + PERFORMANCE_WARNINGS_Y_OFFSET );
		}

		if( m_skippedImages > 0 )
		{
			dc.DrawText( wxString::Format( wxT("%8lu image(s) skipped as the display engine was still busy"), m_skippedImages ), bmpXOff, bmpYOff + SKIPPED_IMAGE_MESSAGE_Y_OFFSET );
		}
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::DeleteAOIs( void )
//-----------------------------------------------------------------------------
{
	AOISet::iterator it = m_aoiSet.begin();
	while( it != m_aoiSet.end() )
	{
		delete *it;
		++it;
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::IncreaseSkippedCounter( unsigned long count )
//-----------------------------------------------------------------------------
{
	m_skippedImages += count;
}

//-----------------------------------------------------------------------------
void ImageCanvas::Init( const wxWindow* const pApp )
//-----------------------------------------------------------------------------
{
	m_boScaleToClientSize = false;
	m_boShowOverlayMessages = false;
	m_pAppEventHandler = pApp->GetEventHandler();
	m_pAppWinID = pApp->GetId();
	m_pIB = 0;
	m_currentZoomFactor = 1.0;
	m_zoomFactor_Max = 1.0;
	m_skippedImages = 0;
}

//-----------------------------------------------------------------------------
void ImageCanvas::OnDblClick( wxMouseEvent& )
//-----------------------------------------------------------------------------
{
	wxCommandEvent e(toggleDisplayArea, m_pAppWinID);
	wxPostEvent( m_pAppEventHandler, e );
}

//-----------------------------------------------------------------------------
void ImageCanvas::OnKeyDown( wxKeyEvent& e )
//-----------------------------------------------------------------------------
{
	if( ( ( e.GetKeyCode() == wxT('+') ) || ( e.GetKeyCode() == WXK_ADD ) || ( e.GetKeyCode() == WXK_NUMPAD_ADD ) )
		&& ( m_currentZoomFactor < m_zoomFactor_Max ) )
	{
		m_currentZoomFactor *= 2.0;
	}
	else if( ( ( e.GetKeyCode() == wxT('-') ) || ( e.GetKeyCode() == WXK_SUBTRACT ) || ( e.GetKeyCode() == WXK_NUMPAD_SUBTRACT ) )
		&& ( m_currentZoomFactor >= 2.0 ) )
	{
		m_currentZoomFactor /= 2.0;
	}
	else
	{
		return;
	}

	RefreshScrollbars();
	if( !m_boScaleToClientSize )
	{
		Refresh( true );
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::RefreshScrollbars( void )
//-----------------------------------------------------------------------------
{
	if( m_pIB && !m_boScaleToClientSize )
	{
		int clientWidth, clientHeight;
		GetClientSize( &clientWidth, &clientHeight );
		int xScrollUnits = ( clientWidth < m_pIB->iWidth * m_currentZoomFactor ) ? clientWidth + m_pIB->iWidth - ( clientWidth / static_cast<int>(m_currentZoomFactor) ) : clientWidth;
		int yScrollUnits = ( clientHeight < m_pIB->iHeight * m_currentZoomFactor ) ? clientHeight + m_pIB->iHeight - ( clientHeight / static_cast<int>(m_currentZoomFactor) ) : clientHeight;
		SetScrollbars( 1, 1, xScrollUnits, yScrollUnits );
	}
	else
	{
		SetScrollbars( 1, 1, 0, 0 );
	}
}

//-----------------------------------------------------------------------------
AOI* ImageCanvas::RegisterAOI( const wxRect& r, const wxColour& c, const wxString& d )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker(m_critSect);
	AOI* pAOI = new AOI(r, c, d);
	m_aoiSet.insert(pAOI);
	return pAOI;
}

//-----------------------------------------------------------------------------
bool ImageCanvas::RemoveAOI( AOI* pAOI )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker(m_critSect);
	AOISet::iterator it = m_aoiSet.find(pAOI);
	if( it != m_aoiSet.end() )
	{
		delete *it;
		m_aoiSet.erase(it);
		return true;
	}
	return false;
}

//-----------------------------------------------------------------------------
bool ImageCanvas::SetAOI( AOI* pAOI, int x, int y, int w, int h )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker(m_critSect);
	AOISet::iterator it = m_aoiSet.find( pAOI );
	if( it != m_aoiSet.end() )
	{
		if( x >= 0 )
		{
			pAOI->m_rect.SetX(x);
		}
		if( y >= 0 )
		{
			pAOI->m_rect.SetY(y);
		}
		if( w >= 0 )
		{
			pAOI->m_rect.SetWidth(w);
		}
		if( h >= 0 )
		{
			pAOI->m_rect.SetHeight(h);
		}
		Refresh( false );
		return true;
	}
	return false;
}

//-----------------------------------------------------------------------------
void ImageCanvas::SetMaxZoomFactor( const mvIMPACT::acquire::ImageBuffer* pIB, int oldMaxDim )
//-----------------------------------------------------------------------------
{
	if( !pIB )
	{
		m_zoomFactor_Max = 1.0;
		return;
	}

	int newMaxDim = std::max( pIB->iWidth, pIB->iHeight );
	if( oldMaxDim != newMaxDim )
	{
		m_zoomFactor_Max = 1.0;
		while( ( ( newMaxDim = newMaxDim >> 1 ) != 0 ) && ( m_currentZoomFactor <= 512.0 ) )
		{
			m_zoomFactor_Max *= 2;
		}
		if( m_currentZoomFactor > m_zoomFactor_Max )
		{
			m_currentZoomFactor = m_zoomFactor_Max;
		}
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::SetScaling( bool boOn )
//-----------------------------------------------------------------------------
{
	if( m_boScaleToClientSize != boOn )
	{
		m_boScaleToClientSize = boOn;
		RefreshScrollbars();
		Refresh( true );
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::SetInfoOverlay( const std::vector<wxString>& infoStrings )
//-----------------------------------------------------------------------------
{
	if( m_infoStringsOverlay != infoStrings )
	{
		m_infoStringsOverlay = infoStrings;
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::SetInfoOverlayMode( bool boOn )
//-----------------------------------------------------------------------------
{
	if( m_boShowInfoOverlay != boOn )
	{
		m_boShowInfoOverlay = boOn;
		Refresh( true );
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::SetPerformanceWarningOutput( bool boOn )
//-----------------------------------------------------------------------------
{
	if( m_boShowOverlayMessages != boOn )
	{
		m_boShowOverlayMessages = boOn;
		Refresh( true );
	}
}

//-----------------------------------------------------------------------------
ImageCanvas::TSaveResult ImageCanvas::StoreImage( const wxImage& img, const wxString& filenameAndPath, const wxString& extension ) const
//-----------------------------------------------------------------------------
{
	if( extension == wxT("bmp") )
	{
		if( img.SaveFile( filenameAndPath, wxBITMAP_TYPE_BMP ) )
		{
			return srOK;
		}
	}
	else if( extension == wxT("png") )
	{
		if( img.SaveFile( filenameAndPath, wxBITMAP_TYPE_PNG ) )
		{
			return srOK;
		}
	}
	return srFailedToSave;
}
